import torch
import numpy as np
import time

def tsv2roll(tsv, audio_length, sample_rate, hop_size, max_midi, min_midi):
    """
    Converting a tsv file into a piano roll
    
    Parameters
    ----------
    tsv : numpy.ndarray
        The tsv label in the array format
    hop_size: int
        Hop size for the spectrogram. It will be used to convert tsvs into piano rolls
    max_midi: int
        The top bin of the pianoroll. Default 108 which corresponds to C8, the highest key on a standard 88-key piano
    min_midi: int
        The lowest bin of the pianoroll. Default 21 which corresponds to A0, the lowest key on a standard 88-key piano
    """
    
    
    n_keys = max_midi - min_midi + 1 # Calutate number of bins for the piano roll
    n_steps = (audio_length - 1) // hop_size + 1 # Calulate number of timesteps for the piano roll
    
    pianoroll = torch.zeros((n_steps, n_keys), dtype=int)
    velocity_roll = torch.zeros((n_steps, n_keys), dtype=int)
    
    for onset, offset, note, vel in tsv:
        left = int(round(onset * sample_rate / hop_size)) # Convert time to time step
        onset_right = min(n_steps, left + 1) # Ensure the time step of onset would not exceed the last time step
        frame_right = int(round(offset * sample_rate / hop_size))
        frame_right = min(n_steps, frame_right) # Ensure the time step of frame would not exceed the last time step
        offset_right = min(n_steps, frame_right + 1)

        f = int(note) - min_midi
        pianoroll[left:onset_right, f] = 3 # assigning onset
        pianoroll[onset_right:frame_right, f] = 2 # assigning sustain
        pianoroll[frame_right:offset_right, f] = 1 # assigning offset
        velocity_roll[left:frame_right, f] = vel    
        
        
        
    return pianoroll, velocity_roll




def check_md5(path, md5_hash):
    with open(path, 'rb') as file_to_check:
        # read contents of the file
        data = file_to_check.read()    
        # pipe contents of the file through
        md5_returned = hashlib.md5(data).hexdigest()

        assert md5_returned==md5_hash, f"{os.path.basename(path)} is corrupted, please download it again"


from torch.utils.data import Dataset
import torchaudio
from pathlib import Path
from typing import Tuple, Union
from torch import Tensor
import torch
import os
import time
import tqdm
import shutil
import glob
import multiprocessing as mp
import warnings
from distutils.dir_util import copy_tree
from torchaudio.compliance import kaldi # for downsampling
from torchaudio.datasets.utils import (
    download_url,
    extract_archive,
)
import hashlib
import torch.nn.functional as F


_CHECKSUMS = {"mls_english_opus": "60390221eec6f456611563b37f0b052c",
              "mls_german_opus":  "b24a9dfd3a8dd1aeabc1341982cc4775",
              "mls_dutch_opus":   "96658c55ef85993a56cf2efbf6f83f57",
              "mls_french_opus":  "b4716b2fd4f67b07c17c6b2a6068641f",
              "mls_spanish_opus": "a2c4b0ff02a71f10ddb21beffbe640ce",
              "mls_italian_opus": "ca5a74d7e97cc62635022719e0ef529d",
              "mls_portuguese_opus": "4dbd6cbdda61268e5d26c4117b0bf769",
              "mls_polish_opus":  "21f83647876c61566c96fdc6298a7b65",
             
              "mls_english": "9d4249911e318c2b8dcfcfecb484d865",
              "mls_german": "91ac982bf63869307f1b8950dfb7c776",
              "mls_dutch": "6b171a16baff0108efd320b1ad65b9d1",
              "mls_french": "4172e807697259bff9ad63661aecabf6",
              "mls_spanish": "6c34698dd522dde28fdc43309e9cc1ac",
              "mls_italian": "dc77f5805aecc7182aa20786032e5dc1",
              "mls_portuguese": "12d54613fae75ae5fb1d55836408f3ee",
              "mls_polish": "ce1a1278006cc373c9d1cb6dbfc03d47"}


espeak_map = {"mls_english_opus": "en",
              "mls_german_opus":  "de",
              "mls_dutch_opus":   "nl",
              "mls_french_opus":  "fr-fr",
              "mls_spanish_opus": "es",
              "mls_italian_opus": "it",
              "mls_portuguese_opus": "pt-pt",
              "mls_polish_opus":  "pl",
              "mls_english": "en",
              "mls_german":  "de",
              "mls_dutch":   "nl",
              "mls_french":  "fr",
              "mls_spanish": "es",
              "mls_italian": "it",
              "mls_portuguese": "pt-pt",
              "mls_polish":  "pl"}
    
def check_md5(path, md5_hash):
    with open(path, 'rb') as file_to_check:
        # read contents of the file
        data = file_to_check.read()    
        # pipe contents of the file through
        md5_returned = hashlib.md5(data).hexdigest()

        assert md5_returned==md5_hash, f"{os.path.basename(path)} is corrupted, please download it again"



class MultilingualLibriSpeech(Dataset):
    """Dataset class for Multilingual LibriSpeech (MLS) dataset.

    Args:
        root (str or Path): Path to the directory where the dataset is found or downloaded.
        
        language_name (str, optional): 
            Choose the folder corresponding to the language,
            for example ``"mls_english"``. For ``"opus"`` version, simply use ``"mls_english_opus".
            ``(default: ``"mls_german_opus"``).
            
        split (str):
            Choose different dataset splits such as ``"train"``, ``"dev"``, and
            ``"test"``. (default: ``"train"``)

        low_resource: bool = False,
        
        download (bool, optional):
            Whether to download the dataset if it is not found at root path. (default: ``False``).
            
        _ext_txt (str, optional):
            Change the file extensions to choose the label type.
            `trans.txt` corresponds to text labels.
            `.ipa_trans.txt` corresponds to IPA labels.
        
    """
    

    def __init__(self,
                 root: Union[str, Path],
                 language_name: str = "mls_italian_opus",
                 split: str = "train",                 
                 low_resource: bool = False,
                 one_hr=0,
                 use_cache: bool = False,
                 refresh: bool = False,
                 sample_rate: int = None,
                 download: bool = False,
                 IPA: bool = False,
                 _ext_txt='.trans.txt'):
        if 'opus' in language_name:
            self._ext_audio = ".opus"
        else:
            self._ext_audio = ".flac"
        self.espeak_map = espeak_map[language_name]
        print(f'using {self.espeak_map} as espeak language')
        self._ext_pytorch = ".pt"
        self._ext_txt = _ext_txt
        self.use_cache = use_cache
        ext_archive = '.tar.gz'
        self.sample_rate = sample_rate
        url = f"https://dl.fbaipublicfiles.com/mls/{language_name}{ext_archive}"
        
        # Getting audio path
        download_path = os.path.join(root, language_name)
        self.download_path = download_path
        self._path = os.path.join(root, language_name, split)
        
        if download:
    
            if os.path.isdir(download_path):
                print(f'Dataset folder exists, skipping download...')
                decision = input(f"Do you want to extract {language_name+ext_archive} again?\n"+
                                 f"This action will overwrite exsiting files, do you still want to continue? [yes/no]")                
            else:
                decision='yes'
                checksum = _CHECKSUMS.get(language_name, None)        
                if not os.path.isdir(root):
                    print(f'Creating download path = {root}')
                    os.makedirs(os.path.join(root))
#                 if os.path.isfile(download_path+ext_archive):
#                     print(f'.tar.gz file exists, proceed to extraction...')
#                 else:
                if os.path.isfile(os.path.join(root, language_name+ext_archive)):
                    print(f'{download_path+ext_archive} already exists, proceed to extraction...')
                else:
                    print(f'downloading...')
                    download_url(url, root, hash_value=checksum, hash_type='md5')
                    
            if decision.lower()=='yes':
                print(f'extracting...')
                extract_archive(download_path+ext_archive)


            print(f'Splitting utterance labels, it might take a long time for large languages such as English.')
#             thread_num = input(f"How many threads do you want to use?\n"+
#                                f"[If you want to use single-thread, enter 0]")
#             thread_num = int(thread_num)
            self.extract_labels('train', num_threads=0, IPA=False)
            self.extract_labels('dev', num_threads=0, IPA=False)
            self.extract_labels('test', num_threads=0, IPA=False)
        
        if os.path.isdir(self._path):
            pass
        else:
            raise FileNotFoundError(f"Dataset not found at {self._path}, please specify the correct location or set `download=True`")



        if low_resource:
            if one_hr==False and isinstance(one_hr, bool):
                print(f'Using 9hr data at {self._path}')
                with open(os.path.join(self._path, 'limited_supervision', '9hr', 'handles.txt'), 'r') as f:
                    self._walker = f.read().splitlines()
            else:
                print(f'Using 1hr data at {self._path}')
                with open(os.path.join(self._path, 'limited_supervision', '1hr', str(one_hr), 'handles.txt'), 'r') as f:
                    self._walker = f.read().splitlines()                
                    
        else:
            if one_hr:
                raise ValueError(f"When `low_resource`={low_resource}, "\
                                 f"`one_hr` should also be False. But received {one_hr} instead")
            else:
                print(f'Using all data at {self._path}')
                # It seems we don't need to + ._ext_audio
                self._walker = sorted(str(p.stem) for p in Path(self._path).glob('audio/*/*/*' + self._ext_audio))
                
        if refresh==True:
            if self.sample_rate:
                print(f'Resampling audio files to {self.sample_rate}Hz. '
                      f'You may change the `sample_rate` argument in the dataset class if you want')
            else:
                print(f'No sample_rate is given, audio files will not be resampled')                
            self.clear_cache()

    def __getitem__(self, n):
        """Load the n-th sample from the dataset.

        Args:
            n (int): The index of the sample to be loaded

        Returns:
            tuple: ``(waveform, sample_rate, utterance, speaker_id, chapter_id, utterance_id)``
        """
        fileid = self._walker[n]
        return self.load_librispeech_item(fileid, self._path, self._ext_audio)
        

    def extract_limited_train_set(self):
        root = os.path.join(self.download_path, 'train', 'limited_supervision')
        audio_root = os.path.join(self.download_path, 'train', 'audio')
        target_root = os.path.join(self.download_path, 'limited_train', 'audio')
        
        if os.path.exists(target_root):
            decision = input(f"{target_root} already exists, this function will overwrite existing files.\n"+
                             f"Do you still want to continue? [yes/no]")
            if decision.lower() == 'no':
                print(f"Audio extraction aborted.")
                return
            elif decision.lower() == 'yes':
                pass
            else:
                raise ValueError(f"'{decision}' is not a valid answer")
        
        # read the files for the 9hr set
        with open(os.path.join(root, '9hr', 'handles.txt')) as f:
            limited_9hr_paths = f.read().splitlines()
        # read the files for the 1hr set
        limited_1hr_paths = {} # create a dictionary to sort the k-fold 1hr set
        for folder_name in range(6):
            # read the files for the 1hr set
            with open(os.path.join(root, '1hr', str(folder_name), 'handles.txt')) as f:
                limited_1hr_paths[folder_name] = f.read().splitlines()

        # Appending 9hr and 1hr set
        unique_paths = set()
        # Add 9hr sets to the unique set
        unique_paths.update(set(limited_9hr_paths))
        # Add the k-fold of 1hr set to the unique set
        for key, values in limited_1hr_paths.items():
            unique_paths.update(set(values))
        
        # Moving the audio files from limited set into a new folder
        for i in tqdm.tqdm(unique_paths, desc='Creating `limited_train` set'):
            speaker_id, chapter_id, utterance_id = i.split('_')
            audio_path = os.path.join(audio_root, speaker_id, chapter_id,
                                      i+self._ext_audio)

            output_path = os.path.join(target_root, speaker_id, chapter_id,
                                       i+self._ext_audio)
            output_folder = os.path.dirname(output_path)
            if not os.path.exists(output_folder):
                os.makedirs(output_folder)
                
            shutil.copy(audio_path, output_path)
            
        # Moving the rest of the files
        train_folder = os.path.join(*audio_root.split('/')[:-1])
        target_train_folder = os.path.join(*target_root.split('/')[:-1])
        # moving metadata inside the `limited_supervision` folder
        copy_tree(os.path.join(train_folder, 'limited_supervision'), os.path.join(target_train_folder, 'limited_supervision'))        
        # moving all .txt files
        for i in glob.glob(os.path.join(train_folder, '*.txt')):
            try:
                shutil.copy(i, target_train_folder)
            except Exception as e:
                pass        
        
    def extract_labels(self, split_name, num_threads=0, Text=True, IPA=False):        
        root = os.path.join(self.download_path, split_name)
        # check if IPA labels already exist
        if IPA:
            ipa_labels = glob.glob(os.path.join(root, 'audio', '*','*','*.ipa_trans.txt'))
            if len(ipa_labels)>0:
                decision = input(f'IPA labels `ipa_trans.txt` in {split_name} folder already exist,'
                                 f'do you want to extract `.ipa_trans.txt` labels again? [yes/no]\n'
                                 f'Warning: All the `.ipa_trans.txt` files inside {split_name} folder will be removed.')
                if decision.lower() == 'no':
                    IPA=False
                elif decision.lower() == 'yes':
                    for i in ipa_labels:
                        os.remove(i)
                else:
                    raise ValueError(f"'{decision}' is not a valid answer")
                    
                    
        labels = glob.glob(os.path.join(root, 'audio', '*','*','*.trans.txt'))
        if len(labels)>0:
            decision = input(f'text labels `.trans.txt` in {split_name} folder already exist,'
                             f'do you want to extract `.trans.txt` labels again? [yes/no]\n'
                             f'Warning: All the `.trans.txt` files inside {split_name} folder will be removed.')
            if decision.lower() == 'no':
                Text=False
                if IPA==False and Text==False:
                    return
            elif decision.lower() == 'yes':
                for i in labels:
                    os.remove(i)
            else:
                raise ValueError(f"'{decision}' is not a valid answer")                    
        
        # Loading all labels into a dictionary
        label_path = os.path.join(root, 'transcripts.txt')
        labels = {}
        with open(label_path,'r') as input_file:
            lines = input_file.readlines()
            for line in tqdm.tqdm(lines, desc='Reading transcripts.txt'):
                utterance_id, text = line.split('\t')
                labels[utterance_id] = text

        # Split labels when the audio exists        
        audio_list = glob.glob(os.path.join(root, 'audio', '*','*',f'*{self._ext_audio}'))
        
        if num_threads==0:
            for audio_file in tqdm.tqdm(audio_list, desc=f"Extracting {split_name} set"):
                self._write_labels(audio_file,labels,root, Text, IPA)
        elif num_threads>0: # use multithread
            pool = mp.Pool(num_threads)
            for audio_file in tqdm.tqdm(audio_list, desc=f"Extracting {split_name} set"):
                r = pool.apply_async(self._write_labels, args=(audio_file,labels,root,Text,IPA,))
    #             r.get()
            pool.close()
            pool.join()
        else:
            raise ValueError(f"'num_threads={num_threads}' is not a valid value")
            
        
    
    def __len__(self) -> int:
        return len(self._walker)
    
    def clear_cache(self):
        """
        This method removes all the .pt file in the dataset.
        By removing all the .pt files, ``__getitem__`` will downsample and
        save the .pt files again when ``use_cache=True``."""
        
        cache_files = glob.glob(os.path.join(self._path,'*','*','*','*.pt'))
        if len(cache_files)>0:
            decision = input(f'{len(cache_files)} .pt files found, confirm clearing cache? [yes/no]')
            if decision.lower()=='yes':
                for i in cache_files:
                    os.remove(i)
            elif decision.lower()=='no':
                print(f'Aborting cache cleaning...')
            else:
                raise ValueError(f'Input {decision} is not recognized, please choose `yes` or `no`.')
        elif len(cache_files)==0:
            print(f'No cache file found')
        else:
            raise ValueError(f'Something is wrong, why does there are {len(cache_files)} cache files')
    
    def load_librispeech_item(self, fileid: str,
                              path: str,
                              ext_audio: str)-> Tuple[Tensor, int, str, int, int, int]:
        speaker_id, chapter_id, utterance_id = fileid.split("_")

        file_text = os.path.join(path, 'audio', speaker_id, chapter_id,
                                 speaker_id+'_'+chapter_id+self._ext_txt)

        saved_name = fileid + self._ext_pytorch
        processed_data_path = os.path.join(path, 'audio', speaker_id, chapter_id, saved_name)
        
        # TODO: check if .pt file exists, if not, move to else
        if self.use_cache==True and os.path.isfile(processed_data_path):
            return torch.load(processed_data_path)
        else:
#             print(f'MLS file not exists')
            file_audio = fileid + ext_audio
            file_audio = os.path.join(path, 'audio', speaker_id, chapter_id, file_audio)


            # Load audio
            start_audioload = time.time()
            waveform, sample_rate = torchaudio.load(file_audio)
            if sample_rate!=self.sample_rate and self.sample_rate!=None: # If the sample_rate is above 16k, downsample it
                waveform = kaldi.resample_waveform(waveform, sample_rate, self.sample_rate)
                sample_rate=self.sample_rate
                
            # Load text
            with open(file_text) as ft:
                for line in ft:
                    fileid_text, utterance = line.strip().split("\t", 1)
                    if '_' in utterance:
                        print(f'file={file_audio} contains _')                    
                    if fileid == fileid_text:
                        break
                else:
                    # Translation not found
                    raise FileNotFoundError("Translation not found for " + fileid_audio)

            batch = {
                     "path": file_audio,
                     "waveform": waveform,
                     "sample_rate": sample_rate,
                     "utterance": utterance,
                     "speaker_id": int(speaker_id),
                     "chapter_id": int(chapter_id),
                     "utterance_id": int(utterance_id)
                    }
            if self.use_cache==True:
                torch.save(batch, processed_data_path)
        return batch
    
    def _write_labels(self, audio_file, labels, root, Text, IPA):
        label_id = os.path.basename(audio_file)[:-5]
        speaker_id, chapter_id, _ = label_id.split('_')
        utterance = labels[label_id]          

        if IPA:
            from phonemizer import phonemize, separator  
            ipa_sequence = phonemize(utterance.lower(),
                                     language=self.espeak_map,
                                     backend='espeak',
                                     strip=True,
                                     language_switch='remove-flags',
                                     separator=separator.Separator(phone=" ", word=" <SPACE> "))

            output_file = os.path.join(root, 'audio', speaker_id, chapter_id,
                                       speaker_id + '_' + chapter_id + '.ipa_trans.txt')
            # writing the ipa labels into each chapter
            ipa_output = label_id + '\t' + ipa_sequence

            with open(output_file, 'a') as f:    
                f.write(ipa_output + '\n')


        if Text:
            # writing the text labels into each chapter        
            output_file = os.path.join(root, 'audio', speaker_id, chapter_id,
                                       speaker_id + '_' + chapter_id + '.trans.txt')

            txt_output = label_id + '\t' + utterance
            with open(output_file, 'a') as f:    
                f.write(txt_output)   


from torch.utils.data import Dataset
import torchaudio
from pathlib import Path
from typing import Tuple, Union
from torch import Tensor
import torch
import os
import time
import tqdm
import shutil
import glob
import multiprocessing as mp
import warnings
from distutils.dir_util import copy_tree
from torchaudio.compliance import kaldi # for downsampling
from torchaudio.datasets.utils import (
    download_url,
    extract_archive,
)
import hashlib
import torch.nn.functional as F            
from AudioLoader.music.utils import check_md5
             
class TIMIT(Dataset):
    """Dataset class for Multilingual LibriSpeech (MLS) dataset.

    Args:
        root (str or Path): Path to the directory where the dataset is found or downloaded.
        
        language_name (str, optional): 
            Choose the folder corresponding to the language,
            for example ``"mls_english"``. For ``"opus"`` version, simply use ``"mls_english_opus".
            ``(default: ``"mls_german_opus"``).
            
        split (str):
            Choose different dataset splits such as ``"train"``, or ``"test"``. (default: ``"train"``).
            ``"dev"`` is not given in this dataset, you have to split part of the ``"train"`` into the ``"dev"`` set.

        
        download (bool, optional):
            Whether to download the dataset if it is not found at root path. (default: ``False``).
            
        _ext_txt (str, optional):
            Change the file extensions to choose the label type.
            `trans.txt` corresponds to text labels.
            `.ipa_trans.txt` corresponds to IPA labels.
        
    """
    

    def __init__(self,
                 root: Union[str, Path],
                 split: str = "train",                 
                 groups: Union[str, int] = 'all',
                 download: bool = False
                ):

        self._ext_audio = ".wav" # The audio format in this dataset is .wav
        ext_archive = '.zip'
#         url = "https://www.kaggle.com/mfekadu/darpa-timit-acousticphonetic-continuous-speech/download"
        url = 'https://data.deepai.org/timit.zip'
        
        # Getting audio path
        archive_name = 'timit'
        folder_name = 'TIMIT'
        download_path = os.path.join(root, folder_name)
        self.download_path = download_path
        assert split.upper()=="TRAIN" or split.upper()=="TEST", f"split={split} is not present in this dataset"
        self._path = os.path.join(root, folder_name, 'data', split.upper())
        
        self.groups = self.available_groups(groups)
        
        checksum = '5b736303c55cf4970926bb9978b655fe'
        
        if download:
            if os.path.isdir(download_path) and os.path.isdir(os.path.join(download_path, 'data')):
                print(f'Dataset folder exists, skipping download...')
                decision = input(f"Do you want to extract {archive_name+ext_archive} again? "
                                 f"To avoid this prompt, set `download=False`\n"
                                 f"This action will overwrite exsiting files, do you still want to continue? [yes/no]") 
                if decision.lower()=='yes':
                    print(f'extracting...')
                    extract_archive(os.path.join(download_path, archive_name+ext_archive))                
            elif os.path.isfile(os.path.join(download_path, 'timit.zip')):
                print(f'timit.zip exists, extracting...')
                check_md5(os.path.join(download_path, archive_name+ext_archive), checksum)
                extract_archive(os.path.join(download_path, archive_name+ext_archive))
            else:
                decision='yes'       
                if not os.path.isdir(download_path):
                    print(f'Creating download path = {root}')
                    os.makedirs(os.path.join(download_path))
#                 if os.path.isfile(download_path+ext_archive):
#                     print(f'.tar.gz file exists, proceed to extraction...')
#                 else:
                if os.path.isfile(os.path.join(download_path, archive_name+ext_archive)):
                    print(f'{download_path+ext_archive} already exists, proceed to extraction...')
                else:
                    print(f'downloading...')
                    try:
                        download_url(url, download_path, hash_value=checksum, hash_type='md5')
                    except:
                        raise Exception('Auto download fails. '+
                                        'You may want to download it manually from:\n'+
                                        url+ '\n' +
                                        f'Then, put it inside {download_path}')

        
        if os.path.isdir(self._path):
            pass
        elif os.path.isfile(os.path.join(download_path, 'timit.zip')):
            print(f'timit.zip exists, extracting...')
            check_md5(os.path.join(download_path, archive_name+ext_archive), checksum)
            extract_archive(os.path.join(download_path, archive_name+ext_archive))
            
        else:
            raise FileNotFoundError(f"Dataset not found at {self._path}, please specify the correct location or set `download=True`")
                    
        print(f'Using all data at {self._path}')
        # It seems we don't need to + ._ext_audio
        self._walker = []
        for group in self.groups:
            for file in [str(p) for p in Path(self._path).glob(os.path.join(group, '*', '*.wav'))]:
                self._walker.append(file)
                

    def __getitem__(self, n):
        """Load the n-th sample from the dataset.

        Args:
            n (int): The index of the sample to be loaded

        Returns:
            tuple: ``(waveform, sample_rate, utterance, speaker_id, chapter_id, utterance_id)``
        """
        file_path = self._walker[n]  
        
        # retreiving DR, gender, speakerid and wav_name
        path_split = file_path.split(os.path.sep)
        DR, speaker, wav_name = path_split[-3:]
        gender = speaker[0]
        speaker_id = speaker[1:]
        
        waveform, sample_rate = torchaudio.load(file_path)
        
        phonemics = self.read_labels(file_path, 'PHN')
        words = self.read_labels(file_path, 'WRD')
        
        batch = {'path': file_path,
                 'waveform': waveform,
                 'sample_rate': sample_rate,
                 'DR': DR,
                 'gender': gender,
                 'speaker_id': speaker_id,
                 'phonemics': phonemics,
                 'words': words
                }
        
        return batch

    
    def read_labels(self, file_path, ext):
        """
        ext: either `'WRD` for word labels or `PHN` for phone labels
        """
        label_path = file_path.replace('WAV.wav', ext)
        labels = ''
        with open(label_path, 'r') as f:
            lines = f.read().splitlines()
            for i in lines:
                labels += f"{(i.split(' ')[-1])} "
            if ext=='PHN':
                # Remove the slience indicator h# at the beginning and the end
                labels = labels.replace('h# ', '')
                
        return labels
    
    def __len__(self) -> int:
        return len(self._walker)
    
            
    def available_groups(self, groups):
        if groups=='all':
            return [f'DR{i+1}' for i in range(9)] # select all dialect regions
        elif isinstance(groups, list):
            return [f'DR{i}' for i in groups]   


import os
from pathlib import Path
from glob import glob
import csv
import shutil
import sys
import pickle
import numpy as np
import random
import musdb
import yaml
from typing import Optional, Callable
# import soundfile
from tqdm import tqdm
import torch
import torch.nn.functional as F
from torch.utils.data import Dataset
import torchaudio
import hashlib
from torchaudio.datasets.utils import (
    download_url,
    extract_archive,
)

from collections import OrderedDict
import math

class FastMUSDB(Dataset):
    def __init__(self,
                 root=None,
                 subsets=['train', 'test'],
                 split=None,
                 seq_duration=6.0,
                 samples_per_track=64,
                 random=False,
                 random_track_mix=False,
                 transform: Optional[Callable] = None
                 ):
        self.root = os.path.expanduser(root)
        self.seq_duration = seq_duration
        self.subsets = subsets
        self.sr = 44100
        self.segment = int(self.seq_duration * self.sr)
        self.split = split
        self.samples_per_track = samples_per_track
        self.random_track_mix = random_track_mix
        self.random = random
        self.sources = ['drums', 'bass', 'other', 'vocals']

        self.transform = transform

        setup_path = os.path.join(
            musdb.__path__[0], 'configs', 'mus.yaml'
        )
        with open(setup_path, 'r') as f:
            self.setup = yaml.safe_load(f)

        self.tracks, self.track_lenghts = self.load_mus_tracks(
            self.sr, self.subsets, self.split)

        if self.seq_duration <= 0:
            self._size = len(self.tracks)
        elif self.random:
            self._size = len(self.tracks) * self.samples_per_track
        else:
            chunks = [l // self.segment for l in self.track_lenghts]
            cum_chunks = np.cumsum(chunks)
            self.cum_chunks = cum_chunks
            self._size = cum_chunks[-1]

    def load_mus_tracks(self, sr, subsets=None, split=None):
        if subsets is not None:
            if isinstance(subsets, str):
                subsets = [subsets]
        else:
            subsets = ['train', 'test']

        if subsets != ['train'] and split is not None:
            raise RuntimeError(
                "Subset has to set to `train` when split is used")

        print("Gathering files ...")
        tracks = []
        track_lengths = []
        for subset in subsets:
            subset_folder = os.path.join(self.root, subset)
            for _, folders, _ in tqdm(os.walk(subset_folder)):
                # parse pcm tracks and sort by name
                for track_name in sorted(folders):
                    if subset == 'train':
                        if split == 'train' and track_name in self.setup['validation_tracks']:
                            continue
                        elif split == 'valid' and track_name not in self.setup['validation_tracks']:
                            continue

                    track_folder = os.path.join(subset_folder, track_name)
                    # add track to list of tracks
                    tracks.append(track_folder)

                    meta = torchaudio.info(os.path.join(
                        track_folder, 'mixture.wav'))
                    assert meta.sample_rate == sr
                    track_lengths.append(meta.num_frames)

        return tracks, track_lengths

    def __len__(self):
        return self._size

    def _get_random_track_idx(self):
        return random.randrange(len(self.tracks))

    def _get_random_start(self, length):
        return random.randrange(length - self.segment + 1)

    def _get_track_from_chunk(self, index):
        track_idx = np.digitize(index, self.cum_chunks)
        if track_idx > 0:
            chunk_start = (index - self.cum_chunks[track_idx - 1]) * self.segment
        else:
            chunk_start = index * self.segment
        return self.tracks[track_idx], chunk_start

    def __getitem__(self, index):
        stems = []
        if self.seq_duration <= 0:
            folder_name = self.tracks[index]
            x, _ = torchaudio.load(
                os.path.join(folder_name, 'mixture.wav')
            )
            for s in self.sources:
                source_name = os.path.join(folder_name, s + '.wav')
                audio, _ = torchaudio.load(source_name)
                stems.append(audio)
        else:
            if self.random:
                track_idx = index // self.samples_per_track
                folder_name, chunk_start = self.tracks[track_idx], self._get_random_start(
                    self.track_lenghts[track_idx])
            else:
                folder_name, chunk_start = self._get_track_from_chunk(index)
            for s in self.sources:
                if self.random_track_mix and self.random:
                    track_idx = self._get_random_track_idx()
                    folder_name, chunk_start = self.tracks[track_idx], self._get_random_start(
                        self.track_lenghts[track_idx])
                source_name = os.path.join(folder_name, s + '.wav')
                audio, _ = torchaudio.load(
                    source_name, frame_offset=chunk_start, num_frames=self.segment
                )
                if audio.shape[1] < self.segment:
                    audio = F.pad(audio.unsqueeze(
                        1), (0, self.segment - audio.shape[1])).squeeze(1)
                stems.append(audio)
            if self.random_track_mix and self.random:
                x = sum(stems)
            else:
                x, _ = torchaudio.load(
                    os.path.join(folder_name, 'mixture.wav'), frame_offset=chunk_start, num_frames=self.segment)
                if x.shape[1] < self.segment:
                    x = F.pad(x.unsqueeze(1),
                              (0, self.segment - x.shape[1])).squeeze(1)

        y = torch.stack(stems)
        if self.transform is not None:
            y = self.transform(y)
            x = y.sum(0)
        return x, y


import os
from pathlib import Path
from glob import glob
import shutil
import sys
import pickle
import numpy as np
import random
from typing import Optional, Callable
# import soundfile
from tqdm import tqdm
import multiprocessing
from joblib import Parallel, delayed
from .utils import tsv2roll, check_md5, files, process_midi, process_csv
import torch
from torch.utils.data import Dataset
import warnings

import torchaudio
from torchaudio.datasets.utils import (
    download_url,
    extract_archive,
)

from collections import OrderedDict
import math

import json
import mido
import importlib

"""
This file is based on https://github.com/jongwook/onsets-and-frames

TODA:
1. Don't load everything on RAM, load only the files when needed (or only do it with MAESTRO?)
2. 
"""

class AMTDataset(Dataset):
    def __init__(self,
                 use_cache=True,
                 download=True,
                 preload=False,
                 sequence_length=None,
                 seed=42,
                 hop_length=512,
                 max_midi=108,
                 min_midi=21,
                 ext_audio='.wav'):
        
        self.use_cache = use_cache
        self.download = download
        self.preload=preload
        self.sequence_length = sequence_length
        self.random = np.random.RandomState(seed)
        self.hop_length = hop_length
        self.max_midi = max_midi
        self.min_midi = min_midi
        self.ext_audio = ext_audio
    
    
    def load(self, index):
        """
        load an audio track and the corresponding labels
        Returns
        -------
            A dictionary containing the following data:
            path: str
                the path to the audio file
            audio: torch.ShortTensor, shape = [num_samples]
                the raw waveform
            label: torch.ByteTensor, shape = [num_steps, midi_bins]
                a matrix that contains the onset/offset/frame labels encoded as:
                3 = onset, 2 = frames after onset, 1 = offset, 0 = all else
            velocity: torch.ByteTensor, shape = [num_steps, midi_bins]
                a matrix that contains MIDI velocity values at the frame locations
        """
        if self.dataset == 'MAESTRO':
            audio_path,tsv_path = self._walker[index]
        elif self.dataset == 'MusicNet' or self.dataset == 'MAPS':
            audio_path = self._walker[index]
            if self.sampling_rate and (self.sampling_rate != 44100): 
                tsv_path = audio_path.replace('.flac', '.tsv')                
            else:
                tsv_path = audio_path.replace(self.ext_audio, '.tsv')
#             saved_data_path = audio_path.replace(self.ext_audio, '.pt')
#             if os.path.exists(audio_path.replace(self.ext_audio, '.pt')) and self.use_cache==True: 
            # Check if .pt files exist, if so just load the files
#                 return torch.load(saved_data_path)
        # Otherwise, create the .pt files
        waveform, sr = torchaudio.load(audio_path)
        if waveform.dim()==2:
            waveform = waveform.mean(0) # converting a stereo track into a mono track
        audio_length = len(waveform)

        tsv = np.loadtxt(tsv_path, delimiter='\t', skiprows=1)

        pianoroll, velocity_roll = tsv2roll(tsv, audio_length, sr, self.hop_length, max_midi=108, min_midi=21)
        data = dict(path=audio_path,
                    sr=sr,
                    audio=waveform,
                    tsv=tsv,
                    pianoroll=pianoroll,
                    velocity_roll=velocity_roll)
        
#         if self.use_cache: # Only save new cache data in .pt format when use_cache==True
#             torch.save(data, saved_data_path)
        return data      
    
    
    def __getitem__(self, index):
        if self.preload:
            data = self._preloader[index]
            result = self.get_segment(data, self.hop_length, self.sequence_length)
            result['sr'] = data['sr']
            return result            
        else:
            data = self.load(index)
            result = self.get_segment(data, self.hop_length, self.sequence_length)
            result['sr'] = data['sr']
            return result    
    
    
    def get_segment(self, data, hop_size, sequence_length=None, max_midi=108, min_midi=21):
        result = dict(path=data['path'])
        audio_length = len(data['audio'])
        pianoroll = data['pianoroll']
        velocity_roll = data['velocity_roll']
    #     start = time.time()
    #     pianoroll, velocity_roll = tsv2roll(data['tsv'], audio_length, data['sr'], hop_size, max_midi, min_midi)
    #     print(f'tsv2roll time used = {time.time()-start}')

        if sequence_length is not None:
            # slicing audio
            assert (audio_length - sequence_length)>0, \
            f"sequence_length={sequence_length} is longer than the "
            f"audio_length={audio_length}. Please reduce the sequence_length"
            begin = self.random.randint(audio_length - sequence_length)
    #         begin = 1000 # for debugging
            end = begin + sequence_length
            result['audio'] = data['audio'][begin:end]

            # slicing pianoroll
            step_begin = begin // hop_size
            n_steps = sequence_length // hop_size
            
            step_end = step_begin + n_steps
            labels = pianoroll[step_begin:step_end, :]
            result['velocity'] = velocity_roll[step_begin:step_end, :]
        else:
            result['audio'] = data['audio']
            labels = pianoroll
            result['velocity'] = velocity_roll

    #     result['audio'] = result['audio'].float().div_(32768.0) # converting to float by dividing it by 2^15
        result['onset'] = (labels == 3).float()
        result['offset'] = (labels == 1).float()
        result['frame'] = (labels > 1).float()
        result['velocity'] = result['velocity'].float().div_(128.0) # not yet normalized
        # print(f"result['audio'].shape = {result['audio'].shape}")
        # print(f"result['label'].shape = {result['label'].shape}")
        return result            

            
    def downsample_exist(self, output_format='flac'):
        if len(self._walker) == 0:
            return 0

        if self.dataset == 'MAESTRO': 
             for wavfile, tsvfile in tqdm(self._walker, desc=f'checking downsampled files'):
                try:
                    dsampled_audio = wavfile.replace(self.ext_audio, f'.{output_format}')
                    assert os.path.isfile(dsampled_audio), f"{dsampled_audio} is missing"
                except Exception as e:
                    warnings.warn(e.args[0])
                    return False
        elif self.dataset == 'MusicNet' or self.dataset == 'MAPS':        
            for wavfile in tqdm(self._walker, desc=f'checking downsampled files'):
                try:
                    dsampled_audio = wavfile.replace(self.ext_audio, f'.{output_format}')
                    assert os.path.isfile(dsampled_audio), f"{dsampled_audio} is missing"
                except Exception as e:
                    warnings.warn(e.args[0])
                    return False
            
        # if downsampled audio exist, check if the sr is correct.
        _, sr = torchaudio.load(dsampled_audio)               
        try:
            assert sr==self.sampling_rate, f'{dsampled_audio} is having a sampling rate of {sr} instead of {self.sampling_rate}'
        except Exception as e:
            warnings.warn(e.args[0])
            return False
        
     
        return True
#             print()
        
    def clear_caches(self):
        """"Clearing existing .pt files"""
        cache_list = list(Path(os.path.join(self.root, self.name_archive)).rglob('*.pt'))
        decision = input(f"Found {len(cache_list)} .pt files"
                         f"Do you want to remove them?"
                         f"Choosing [no] if you want to double check the list of files to be removed when [yes/no]")
        
        if decision.lower()=='yes':
            for file in cache_list:
                os.remove(file)
        elif decision.lower()=='no':
            return cache_list
        else:
            print(f"[{decision}] is not a supported answer. Clearing skipped.")
            return cache_list            

    def __len__(self):
        return len(self._walker)

    
class MAPS(AMTDataset):
    def __init__(self,
                 root='./',
                 groups='all',
                 data_type='MUS',
                 overlap=True,
                 sampling_rate=None,
                 **kwargs):
        """
        This Dataset inherits from AMTDataset.
        
        Parameters
        ----------        
        root: str
            The folder that contains the MAPS dataset folder

        groups: list or str
            Choose which sub-folders to load. Avaliable choices are 
            `train`, `test`, `all`.
            Default: `all`, which stands for loading all sub-folders.
            Alternatively, users can provide a list of subfolders to be loaded.

        data_type: str
            Four different types of data are available, `MUS`, `ISOL`, `RAND`, `UCHO`.
            Default: `MUS`, which stands for full music pieces.
            
        overlap: bool
            TODO: To control if overlapping songs in the train set to be load.
            Default: False, which means that it will ignore audio clips in the train set
                     which already exist in the test set ('ENSTDkAm' and 'ENSTDkCl')
                     
        use_cache: bool
            If it is set to `True`, the audio, piano roll and its metadata would be saved as a .pt file.
            Loading directly from the .pt files would be slightly faster than loading raw audio and tsv files
            Default:True
            
        download: bool
            To automatically download the dataset if it is set to `True`
            Default: True
            
        preload: bool
            When it is set to `True`, the data will be loaded into RAM, which makes loading faster.
            For large dataset, RAM memory might not be enough to store all the data, and we can
            use `preload=False` to read data on the fly.
            Default: False

        sequence_length: int
            The length of audio segment to be extracted.
            Since the audio is paired with a tsv file,
            changing this will automactially change the piano roll lenght too.
            Default: None

        seed: int
            This seed controls the segmentation indices, which allows the data loading to be reproducible.
            Default: 42
            
        hop_length: int
            It should be the same as the spectrogram hop_length,
            so that the piano roll timesteps will aligned with the spectrograms.
            Default: 512     

        max_midi: int
            The highest MIDI note to be appeared on the piano roll 
            Default: 108, which is equivalent to the highest note, C8, on a piano
            
        min_midi: int
            The lowest MIDI note to be appeared on the piano roll 
            Default: 21, which is equivalent to the lowest note, A0, on a piano  
            
            
        ext_audio: str
            The audio format to load. Most dataset provides audio in the format of `.wav` files.
            The `.resample(sr)` function resamples audio into `.flac` format.
            Therefore, changing the target audio format to be load can be used to control
            which set of audio data to use
            Default: '.wav'            
        """
        
        self.overlap = overlap
        super().__init__(**kwargs)
        
#         self.url = "https://amubox.univ-amu.fr/s/iNG0xc5Td1Nv4rR/download"
#         self.checksum = '02a8f140dc9a7c85639b0c01e5522add'

        self.url_dict = {'AkPnBcht': 'https://amubox.univ-amu.fr/s/iNG0xc5Td1Nv4rR/download?path=%2F&files=AkPnBcht.zip&downloadStartSecret=2qjs7gi2ixw',
                    'AkPnBsdf': 'https://amubox.univ-amu.fr/s/iNG0xc5Td1Nv4rR/download?path=%2F&files=AkPnBsdf.zip&downloadStartSecret=lqjko9sgjv',
                    'AkPnCGdD': 'https://amubox.univ-amu.fr/s/iNG0xc5Td1Nv4rR/download?path=%2F&files=AkPnCGdD.zip&downloadStartSecret=sqfhv2kjb4d',
                    'AkPnStgb': 'https://amubox.univ-amu.fr/s/iNG0xc5Td1Nv4rR/download?path=%2F&files=AkPnStgb.zip&downloadStartSecret=p50p2c8wjka',
                    'ENSTDkAm1': 'https://amubox.univ-amu.fr/s/iNG0xc5Td1Nv4rR/download?path=%2F&files=ENSTDkAm1.zip&downloadStartSecret=4jd7mmoberd',
                    'ENSTDkAm2': 'https://amubox.univ-amu.fr/s/iNG0xc5Td1Nv4rR/download?path=%2F&files=ENSTDkAm2.zip&downloadStartSecret=bfeekv6zios',
                    'ENSTDkCl': 'https://amubox.univ-amu.fr/s/iNG0xc5Td1Nv4rR/download?path=%2F&files=ENSTDkCl.zip&downloadStartSecret=1ekyv85ij5wh',
                    'SptkBGAm': 'https://amubox.univ-amu.fr/s/iNG0xc5Td1Nv4rR/download?path=%2F&files=SptkBGAm.zip&downloadStartSecret=lckjw1lgks',
                    'SptkBGCl': 'https://amubox.univ-amu.fr/s/iNG0xc5Td1Nv4rR/download?path=%2F&files=SptkBGCl.zip&downloadStartSecret=cho91vy3swp',
                    'StbgTGd2': 'https://amubox.univ-amu.fr/s/iNG0xc5Td1Nv4rR/download?path=%2F&files=StbgTGd2.zip&downloadStartSecret=htixu8ryz3h'
                    }

        self.hash_dict = {'AkPnBcht': '44f0b64a7cda143cfb25e217300df743',
                     'AkPnBsdf': '3f26df7b2f104a3df3f27cd7b91c461a',
                     'AkPnCGdD': 'c15622c304e72c0d6223ce9f7070036f',
                     'AkPnStgb': '75bda66fb21b927338ae2c042333175d',
                     'ENSTDkAm1': 'a47136282f92254722dd92aef70c6262',
                     'ENSTDkAm2': '2d56f30b88e71010c66ab8c32cd7f582',
                     'ENSTDkCl': 'd51e503d592136f3c400f832ecb767ce',
                     'SptkBGAm': '6d37344417bebd2353aeeff2b7d8232f',
                     'SptkBGCl': '37d91c67d96c6612b348f171244bfb2c',
                     'StbgTGd2': '2e08fc13143a525d5b316f09ed38f9d4'
                    }        
        
        self.root = root
#         self.ext_archive = '.tar'
        self.ext_archive = '.zip'
        self.name_archive = 'MAPS'
        self.original_ext = '.wav'        
        self.data_type = data_type
        self.sampling_rate = sampling_rate
        self.dataset = 'MAPS'
             
        self.groups = groups if isinstance(groups, list) else self.available_groups(groups)
        
        
        if self.download:
            # Check if MAPS folder exists
            if not os.path.isdir(os.path.join(root, self.name_archive)):
                os.makedirs(os.path.join(root, self.name_archive)) 

            if self._check_all_groups_exist(self.groups): # If data folder does not exist, check if zip files exist
                print(f'All zip files exist.')
                self.extract_subfolders(self.groups)
                # Downsampling audio to 16kHz flac formats
                self.extract_tsv()                   

#         Downloading the complete zip file is broken at "https://amubox.univ-amu.fr/s/iNG0xc5Td1Nv4rR/download"
#         if self.download:
#             if os.path.isdir(os.path.join(self.root, self.name_archive)):
#                 print(f'Dataset folder exists, skipping download...\n'
#                       f'Checking sub-folders...')
#                 self.extract_subfolders(groups)
#                 self.extract_tsv()
#             elif os.path.isfile(os.path.join(self.root, self.name_archive+self.ext_archive)):
#                 print(f'.tar file exists, skipping download...')
#                 print(f'Extracting MAPS.tar')
#                 extract_archive(os.path.join(self.root, self.name_archive+self.ext_archive))
#                 self.extract_subfolders(groups)
#                 self.extract_tsv()                
#             else:
#                 if not os.path.isdir(self.root):
#                     print(f'Creating download path = {self.root}')
#                     os.makedirs(os.path.join(self.root))
                    
#                 print(f'Downloading from {self.url}\n'
#                       f"If download won't start automatically, please visit "
#                       f"https://amubox.univ-amu.fr/index.php/s/iNG0xc5Td1Nv4rR "
#                       f"to download manually")
#                 download_url(self.url, root, hash_value=self.checksum, hash_type='md5')
#                 print(f'Extracting MAPS.tar')
#                 extract_archive(os.path.join(self.root, self.name_archive+self.ext_archive))
#                 self.extract_subfolders(groups)
#                 self.extract_tsv()

        
        else:
            if os.path.isdir(os.path.join(root, self.name_archive)):
                print(f'MAPS folder found, checking content integrity...')
                self.extract_subfolders(self.groups)   
            else:
                raise ValueError(f'{root} does not contain the MAPS folder, '
                                 f'please specify the correct path or download it by setting `download=True`')
                
#         print(f"Loading {len(groups)} group{'s' if len(groups) > 1 else ''} "
#               f"of {self.__class__.__name__} at {os.path.join(self.root, self.name_archive)}")
        self._walker = [] #if sr=none or 44100, self.walker is .wav 
    
        for group in self.groups:
            wav_paths = glob(os.path.join(self.root, self.name_archive, group, data_type, f'*{self.ext_audio}'))
            self._walker.extend(wav_paths)

        if self.overlap==False:
            assert groups!='test', 'When loading test set, please set overlap=True'
    
#             template = pkg_resources.read_text('overlapping.pkl')
#             with open(a, 'rb') as f:
            test_names = pickle.load(importlib.resources.open_binary("AudioLoader.music", 'overlapping.pkl'))
            filtered_flacs = []    
            for i in self._walker:
                if any([substring in i for substring in test_names]):
                    pass
                else:
                    filtered_flacs.append(i)
            self._walker = filtered_flacs 
            
        if self.preload:
            self._preloader = []
            for i in tqdm(range(len(self._walker)),desc=f'Pre-loading data to RAM'):
                self._preloader.append(self.load(i))
                
        if self.sampling_rate and (self.sampling_rate != 44100):
            # When sampling rate is given, it will automatically create a downsampled copy
            if self.downsample_exist('flac'):
                print(f"downsampled audio exists, skipping downsampling")
            else:
                print('doing resample()')
                self.resample(sampling_rate, 'flac', num_threads=4)
                
            for idx, audio in tqdm((enumerate(self._walker))):
                self._walker[idx] = audio.replace('.wav', '.flac')  
            # if need resample, auto change .wav path to .flac path in self._walker
            
            # reload the flac audio after downsampling only when _walker is empty
#             if len(self._walker) == 0:
#                 for group in groups:
#                     wav_paths = glob(os.path.join(self.root, self.name_archive, group, data_type, f'*{self.ext_audio}'))
#                     self._walker.extend(wav_paths)
            
            
         
        print(f'{len(self._walker)} audio files found')
        if self.use_cache:
            print(f'use_cache={self.use_cache}: it will use existing cache files (.pt) and ignore other changes '
                  f'such as ext_audio, max_midi, min_midi, and hop_length.\n'
                  f'Please use .clear_caches() to remove existing .pt files to refresh caches')

    def extract_subfolders(self, groups):
        for group in groups:
            group_path = os.path.join(self.root, self.name_archive, group)
            if not os.path.isdir(group_path):
                print(f'Extracting sub-folder {group}...', end='\r')
                if group=='ENSTDkAm':
                    # ENSTDkAm consists of ENSTDkAm1.zip and ENSTDkAm2.zip
                    # Extract and merge both ENSTDkAm1.zip and ENSTDkAm2.zip as ENSTDkAm
                    check_md5(os.path.join(self.root, self.name_archive, group+'1.zip'), self.hash_dict[group+'1'])
                    check_md5(os.path.join(self.root, self.name_archive, group+'2.zip'), self.hash_dict[group+'2'])
                    extract_archive(os.path.join(self.root, self.name_archive, group+'1.zip'))
                    extract_archive(os.path.join(self.root, self.name_archive, group+'2.zip'))
                else:
                    check_md5(os.path.join(self.root, self.name_archive, group+'.zip'), self.hash_dict[group])
                    extract_archive(os.path.join(self.root, self.name_archive, group+'.zip'))
                print(f' '*50, end='\r')
                print(f'{group} extracted.')        
                
                
    def _check_all_groups_exist(self, groups):
        print("Checking if data folders already exist...")
        for group in groups:
            if os.path.isdir(os.path.join(self.root, self.name_archive, group)):
                pass
            else:
                print(f"{group} not found, proceeding to check if the zip file exists", end='\r')
                if group=='ENSTDkAm': # ENSTDkAm has 2 zip files ENSTDkAm1.zip and ENSTDkAm2.zip
                    zip_list = ['ENSTDkAm1', 'ENSTDkAm2']
                    for group in zip_list:
                        self._check_and_download_zip(group)
                else:
                    self._check_and_download_zip(group)
        return True

    def _check_and_download_zip(self, group):
        if os.path.isfile(os.path.join(self.root, self.name_archive, group+'.zip')):
            print(f"{group+'.zip'} exists" + " "*100)
            pass
        else:
            print(" "*shutil.get_terminal_size().columns, end='\r')
            print(f"{group+'.zip'} not found, proceeding to download")
            download_url(self.url_dict[group],
                         os.path.join(self.root, self.name_archive),
                         hash_value=self.hash_dict[group],
                         hash_type='md5')           
                    
    def extract_tsv(self):
        """
        Convert midi files into tsv files for easy loading.
        """
        
        tsvs = glob(os.path.join(self.root, self.name_archive, '*', self.data_type, '*.tsv'))
        num_tsvs = len(tsvs)
        if num_tsvs>0:
            decision = input(f"There are already {num_tsvs} tsv files.\n"+
                             f"Do you want to overwrite them? [yes/no]")
        elif num_tsvs==0:
            decision='yes'
            
        if decision.lower()=='yes':
            midis = glob(os.path.join(self.root, self.name_archive, '*', self.data_type, '*.mid')) # loading lists of midi    
            Parallel(n_jobs=multiprocessing.cpu_count())\
                    (delayed(process_midi)(in_file, out_file) for in_file, out_file in files(midis, output_dir=False))
                
    def available_groups(self, group):
        if group=='train':
            return ['AkPnBcht', 'AkPnBsdf', 'AkPnCGdD', 'AkPnStgb', 'SptkBGAm', 'SptkBGCl', 'StbgTGd2']
        elif group=='test':
            return ['ENSTDkAm', 'ENSTDkCl']
        elif group=='all':
            return ['AkPnBcht', 'AkPnBsdf', 'AkPnCGdD', 'AkPnStgb', 'ENSTDkAm', 'ENSTDkCl', 'SptkBGAm', 'SptkBGCl', 'StbgTGd2']
        

    def clear_audio(self, audio_format='.flac'):
        clear_list = []
        for group in self.groups:
            audio_paths = glob(os.path.join(self.root,
                                          self.name_archive,
                                          group,
                                          self.data_type,
                                          f'*{audio_format}'))
            clear_list.extend(audio_paths)
            
        num_files = len(clear_list)
        if num_files>0:
            decision = input(f"{num_files} files found, do you want to clear them?[yes/no]")
            if decision.lower()=='yes':
                for i in clear_list:
                    os.remove(i)
            elif decision.lower()=='no':
                print(f'aborting...')
                
    def resample(self, sr, output_format='flac', num_threads=-1):
        """
        ```python
        dataset = MAPS('./Folder', groups='all', ext_audio='.flac')
        dataset.resample(sr, output_format='flac', num_threads=4)
        ```
        It is known that sometimes num_threads>0 (using multiprocessing) might cause corrupted audio after resampling
        
        Resample audio clips to the target sample rate `sr` and the target format `output_format`.
        This method requires `pydub`.
        After resampling, you need to create another instance of `MAPS` in order to load the new
        audio files instead of the original `.wav` files.
        """
        
        from pydub import AudioSegment        
        def _resample(wavfile, sr, output_format):
            sound = AudioSegment.from_wav(wavfile.replace(self.ext_audio, self.original_ext))
            sound = sound.set_frame_rate(sr) # downsample it to sr
            sound = sound.set_channels(1) # Convert Stereo to Mono
            sound.export(wavfile.replace(self.original_ext, f'.{output_format}'), format=output_format)            
            
        if num_threads==-1:    
            Parallel(n_jobs=multiprocessing.cpu_count())\
            (delayed(_resample)(wavfile, sr, output_format)\
             for wavfile in tqdm(self._walker,
                                 desc=f'Resampling to {sr}Hz .{output_format} files'))            
        elif num_threads==0:
            for wavfile in tqdm(self._walker, desc=f'Resampling to {sr}Hz .{output_format} files'):
                _resample(wavfile, sr, output_format)
        else:
            Parallel(n_jobs=num_threads)\
            (delayed(_resample)(wavfile, sr, output_format)\
             for wavfile in tqdm(self._walker,
                                 desc=f'Resampling to {sr}Hz .{output_format} files'))                

                
class MusicNet(AMTDataset):
    def __init__(self,
                 root='./',
                 groups='all',
                 sampling_rate=None,
                 **kwargs):
        """
        root (str): The folder that contains the MusicNet dataset folder
        groups (list or str): Choose which sub-folders to load. Avaliable choices are 
                              `train`, `test`, `all`. Default is `all`, which means loading
                               all sub-folders. Alternatively, users can provide a list of
                               subfolders to be loaded.                
        """
        
        super().__init__(**kwargs)
        
#         self.url = "https://homes.cs.washington.edu/~thickstn/media/musicnet.tar.gz" # old MusicNet link
        self.url = "https://zenodo.org/record/5120004/files/musicnet.tar.gz?download=1"
        self.checksum = 'd41d8cd98f00b204e9800998ecf8427e'
        self.root = root
        self.ext_archive = '.tar.gz'
        self.name_archive = 'musicnet'
        self.original_ext = '.wav'
        self.sampling_rate  = sampling_rate
        self.dataset = 'MusicNet'
             
        groups = groups if isinstance(groups, list) else self.available_groups(groups)
        self.groups = groups

        if self.download:
            if os.path.isdir(os.path.join(self.root, self.name_archive)):
                print(f'{self.name_archive} folder exists, skipping download')
                print(f'Converting csv files into tsv files')
                self.csv2tsv()
            elif os.path.isfile(os.path.join(self.root, self.name_archive+self.ext_archive)):
                print(f'{self.name_archive+self.ext_archive} exists, skipping download')
                print(f'Extracting {self.name_archive+self.ext_archive}')
                extract_archive(os.path.join(self.root, self.name_archive+self.ext_archive))
                print(f'Converting csv files into tsv files')
                self.csv2tsv()                
            else:
                if not os.path.isdir(self.root):
                    print(f'Creating download path = {self.root}')
                    os.makedirs(os.path.join(self.root))
                    
                print(f'Downloading from {self.url}')
                download_url(self.url, root, hash_value=self.checksum, hash_type='md5')
                print(f'Extracting musicnet.tar.gz')
                extract_archive(os.path.join(self.root, self.name_archive+self.ext_archive))
                print(f'Converting csv files into tsv files')
                self.csv2tsv()
        
        else:
            if os.path.isdir(os.path.join(root, self.name_archive)):
                print(f'{self.name_archive} folder found')
            elif os.path.isfile(os.path.join(self.root, self.name_archive+self.ext_archive)):
                print(f'{self.name_archive} folder not found, but {self.name_archive+self.ext_archive} exists')
                print(f'Extracting {self.name_archive+self.ext_archive}')
                extract_archive(os.path.join(self.root, self.name_archive+self.ext_archive))
                print(f'Converting csv files into tsv files')
                self.csv2tsv()                
            else:
                raise ValueError(f'{root} does not contain the MusicNet folder, '
                                 f'please specify the correct path or download it by setting `download=True`')  
                
#         print(f"Loading {len(groups)} group{'s' if len(groups) > 1 else ''} "
#               f"of {self.__class__.__name__} at {os.path.join(self.root, self.name_archive)}")
        self._walker = []
    
        for group in groups:
            wav_paths = glob(os.path.join(self.root, self.name_archive, f'{group}_data', f'*{self.ext_audio}'))
            self._walker.extend(wav_paths)
            
                
        if self.sampling_rate and (self.sampling_rate != 44100):
            # When sampling rate is given, it will automatically create a downsampled copy
            if self.downsample_exist('flac'):
                print(f"downsampled audio exists, skipping downsampling")
            else:
                print('doing resample()')
                self.resample(sampling_rate, 'flac', num_threads=4)
                
            for idx, audio in tqdm((enumerate(self._walker))):
                self._walker[idx] = audio.replace('.wav', '.flac')  
            # if need resample, auto change .wav path to .flac path in self._walker
            
            # reload the flac audio after downsampling only when _walker is empty
            if len(self._walker) == 0:
                for group in groups:
                    wav_paths = glob(os.path.join(self.root, self.name_archive, f'{group}_data', f'*{self.ext_audio}'))
                    self._walker.extend(wav_paths)            
            
        if self.preload:
            self._preloader = []
            for i in tqdm(range(len(self._walker)),desc=f'Pre-loading data to RAM'):
                self._preloader.append(self.load(i))            
                
        print(f'{len(self._walker)} audio files found')
 
                    
    def csv2tsv(self):
        """
        Convert csv files into tsv files for easy loading.
        """
        for group in self.groups:
            tsvs = glob(os.path.join(self.root, self.name_archive, f"{group}_data", '*.tsv'))            
            num_tsvs = len(tsvs)
            if num_tsvs>0:
                decision = input(f"There are already {num_tsvs} tsv files.\n"+
                                 f"Do you want to overwrite them? [yes/no]")
            elif num_tsvs==0:
                decision='yes'

            if decision.lower()=='yes':
                csvs = glob(os.path.join(self.root, self.name_archive, f'{group}_labels','*.csv')) # loading lists of csvs
                Parallel(n_jobs=multiprocessing.cpu_count())\
                        (delayed(process_csv)(in_file, out_file) for in_file, out_file in files(csvs, output_dir=False))

            # moving all tsv files to the data folder where wav files are located
            # This is to make sure we can use the same __get_item__ fuction
            tsvs = glob(os.path.join(self.root, self.name_archive, f"{group}_labels", '*.tsv'))
            for tsv in tsvs:
                target_path = tsv.replace(f"{group}_labels", f"{group}_data")
                shutil.move(tsv, target_path)
                
    def available_groups(self, group):
        if group=='train':
            return ['train']
        elif group=='test':
            return ['test']
        elif group=='all':
            return ['train', 'test']
        
    def resample(self, sr, output_format='flac', num_threads=-1):
        """
        ```python
        dataset = MAPS('./Folder', groups='all', ext_audio='.flac')
        dataset.resample(sr, output_format='flac', num_threads=4)
        ```
        It is known that sometimes num_threads>0 (using multiprocessing) might cause corrupted audio after resampling
        
        Resample audio clips to the target sample rate `sr` and the target format `output_format`.
        This method requires `pydub`.
        After resampling, you need to create another instance of `MAPS` in order to load the new
        audio files instead of the original `.wav` files.
        """      
        
        from pydub import AudioSegment        
        def _resample(wavfile, sr, output_format):
            sound = AudioSegment.from_wav(wavfile.replace(self.ext_audio, self.original_ext))
            sound = sound.set_frame_rate(sr) # downsample it to sr
            sound = sound.set_channels(1) # Convert Stereo to Mono
            sound.export(wavfile.replace(self.original_ext, f'.{output_format}'), format=output_format)            
            
        if num_threads==-1:    
            Parallel(n_jobs=multiprocessing.cpu_count())\
            (delayed(_resample)(wavfile, sr, output_format)\
             for wavfile in tqdm(self._walker,
                                 desc=f'Resampling to {sr}Hz .{output_format} files'))            
        elif num_threads==0:
            for wavfile in tqdm(self._walker, desc=f'Resampling to {sr}Hz .{output_format} files'):
                _resample(wavfile, sr, output_format)
        else:
            Parallel(n_jobs=num_threads)\
            (delayed(_resample)(wavfile, sr, output_format)\
             for wavfile in tqdm(self._walker,
                                 desc=f'Resampling to {sr}Hz .{output_format} files'))  
            
            
            
#start of class MAESTRO
def parse_midi(path):
    """open midi file and return np.array of (onset, offset, note, velocity) rows"""
    midi = mido.MidiFile(path)

    time = 0
    sustain = False
    events = []
    for message in midi:
        time += message.time

        if message.type == 'control_change' and message.control == 64 and (message.value >= 64) != sustain:
            # sustain pedal state has just changed
            sustain = message.value >= 64
            event_type = 'sustain_on' if sustain else 'sustain_off'
            event = dict(index=len(events), time=time, type=event_type, note=None, velocity=0)
            events.append(event)

        if 'note' in message.type:
            # MIDI offsets can be either 'note_off' events or 'note_on' with zero velocity
            velocity = message.velocity if message.type == 'note_on' else 0
            event = dict(index=len(events), time=time, type='note', note=message.note, velocity=velocity, sustain=sustain)
            events.append(event)

    notes = []
    for i, onset in enumerate(events):
        if onset['velocity'] == 0:
            continue

        # find the next note_off message
        offset = next(n for n in events[i + 1:] if n['note'] == onset['note'] or n is events[-1])

        if offset['sustain'] and offset is not events[-1]:
            # if the sustain pedal is active at offset, find when the sustain ends
            offset = next(n for n in events[offset['index'] + 1:] if n['type'] == 'sustain_off' or n is events[-1])

        note = (onset['time'], offset['time'], onset['note'], onset['velocity'])
        notes.append(note)

    return np.array(notes)
    
    
class MAESTRO(AMTDataset):
    def __init__(self, 
                 root, 
                 groups=None,
                 sampling_rate=None,
                 **kwargs):
        super().__init__(**kwargs)
        
        self.url_dict = {'maestro-v2.0.0': 'https://storage.googleapis.com/magentadata/datasets/maestro/v2.0.0/maestro-v2.0.0.zip',
                         'maestro-v2.0.0-midi': 'https://storage.googleapis.com/magentadata/datasets/maestro/v2.0.0/maestro-v2.0.0-midi.zip',
                        }
        self.hash_dict = {'maestro-v2.0.0': '7a6c23536ebcf3f50b1f00ac253886a7',
                          'maestro-v2.0.0-midi': '8a45cc678a8b23cd7bad048b1e9034c5'
                         }
        self.root = root
        self.ext_archive = '.zip'
        self.name_archive = 'maestro-v2.0.0'
        self.midi_name_archive = 'maestro-v2.0.0-midi'
        self.original_ext = '.wav'        
        self.sampling_rate = sampling_rate
        self.dataset = 'MAESTRO'

             
        groups = groups if isinstance(groups, list) else self.available_groups(groups)
        self.groups = groups
        
        if self.download: 
            if os.path.isdir(os.path.join(self.root, self.name_archive)): #check for the maestro-v2.0.0 folder
                print(f'Dataset folder exists, skipping download...')
            elif os.path.isfile(os.path.join(self.root, 'maestro-v2.0.0.zip')) and os.path.isfile(os.path.join(self.root, 'maestro-v2.0.0-midi.zip')): 
                #when don't have 'maestro-v2.0.0' folder, check for the zip and extract
                print(f'maestro-v2.0.0.zip and maestro-v2.0.0-midi.zip exists, checking MD5...')
                check_md5(os.path.join(self.root, self.name_archive+self.ext_archive), self.hash_dict[self.name_archive])
                tqdm(extract_archive(os.path.join(self.root, self.name_archive+self.ext_archive)))
                
                check_md5(os.path.join(self.root, self.midi_name_archive+self.ext_archive), self.hash_dict[self.midi_name_archive])
                tqdm(extract_archive(os.path.join(self.root, self.midi_name_archive+self.ext_archive)))
            
            elif (not os.path.isfile(os.path.join(self.root, 'maestro-v2.0.0.zip'))) and (not os.path.isfile(os.path.join(self.root, 'maestro-v2.0.0-midi.zip'))):
                #download both zip file and extract
                download_url(self.url_dict['maestro-v2.0.0'], self.root, hash_value=self.hash_dict['maestro-v2.0.0'], hash_type='md5')
                download_url(self.url_dict['maestro-v2.0.0-midi'], self.root, hash_value=self.hash_dict['maestro-v2.0.0-midi'], hash_type='md5')
                print(f'Download finished, extracting zip file')
                tqdm(extract_archive(os.path.join(self.root, self.name_archive+self.ext_archive)))
                tqdm(extract_archive(os.path.join(self.root, self.midi_name_archive+self.ext_archive)))
       
        elif os.path.isdir(os.path.join(self.root, self.name_archive)): #when download=False, and maestro-v2.0.0 folder exist
            pass
        elif os.path.isfile(os.path.join(self.root, 'maestro-v2.0.0.zip')) and os.path.isfile(os.path.join(self.root, 'maestro-v2.0.0-midi.zip')): 
            #when don't have 'maestro-v2.0.0' folder, check for the zip 
            print(f'maestro-v2.0.0.zip and maestro-v2.0.0-midi.zip exists, checking MD5...')
            check_md5(os.path.join(self.root, self.name_archive+self.ext_archive), self.hash_dict[self.name_archive])
            extract_archive(os.path.join(self.root, self.name_archive+self.ext_archive))

            check_md5(os.path.join(self.root, self.midi_name_archive+self.ext_archive), self.hash_dict[self.midi_name_archive])
            extract_archive(os.path.join(self.root, self.midi_name_archive+self.ext_archive))

        else:    
            raise FileNotFoundError(f"Dataset not found at {self.root}, please specify the correct location or set `download=True`")

            
        self._walker = []     #self._walker is the combined audio path audio list for all groups    
        for group in self.groups: # self.groups is a list of str, can be more than one group inside
                self._walker.extend(self.files(group)) 
        
        if self.preload:
            self._preloader = []
            for i in tqdm(range(len(self._walker)),desc=f'Pre-loading data to RAM'):
                self._preloader.append(self.load(i))
                
        
        if self.sampling_rate and (self.sampling_rate != 44100):
            # When sampling rate is given, it will automatically create a downsampled copy
            if self.downsample_exist('flac'): #it will return False if sr of existing .flac not match with self.sampling_rate 
                print(f"downsampled audio exists, skipping downsampling")
            else:
                print('doing resample()')
                self.resample(self.sampling_rate, 'flac', num_threads=4)

            
#             # reload the flac audio after downsampling only when _walker is empty
#             if len(self._walker) == 0:
#                 for group in groups:
#                     wav_paths = glob(os.path.join(self.root, self.name_archive, group, f'*{self.ext_audio}'))
#                     self._walker.extend(wav_paths)
    
    def files(self, group):
        metadata = json.load(open(os.path.join(self.root, self.name_archive, 'maestro-v2.0.0.json')))
        file = sorted([(os.path.join(self.root, self.name_archive, row['audio_filename']),
                         os.path.join(self.root, self.name_archive, row['midi_filename'])) for row in metadata if row['split'] == group])

        files=[] #use to create [_walker]
        for audio, midi in file:                
            if (self.sampling_rate is None) or (self.sampling_rate == 44100):
                files.append((audio, midi))
            else:
                self.sampling_rate != 44100 
                files.append((audio.replace('.wav', '.flac'), midi))

        _walker = []  #_walker is the list for individaul group, _walker[idx]= ('path to wav', 'path to tsv')
        for audio_path, midi_path in tqdm(files):          
            tsv_filename = midi_path.replace('.midi', '.tsv').replace('.mid', '.tsv')
            if not os.path.exists(tsv_filename):
                midi = parse_midi(midi_path)
                np.savetxt(tsv_filename, midi, fmt='%.6f', delimiter='\t', header='onset,offset,note,velocity')
            _walker.append((audio_path, tsv_filename))
        return _walker 
 
  
    
    def available_groups(self, group):
        if group=='train':
            return ['train']
        elif group=='test':
            return ['test']
        elif group=='validation':
            return ['validation']
        
    def resample(self, sr, output_format='flac', num_threads=-1):
        """
        It is known that sometimes num_threads>0 (using multiprocessing) might cause corrupted audio after resampling
        
        Resample audio clips to the target sample rate `sr` and the target format `output_format`.
        This method requires `pydub`.
        After resampling, you need to create another instance of `MAESTRO` in order to load the new
        audio files instead of the original `.wav` files.
        """   
        
        from pydub import AudioSegment        
        def _resample(wavfile, sr, output_format):  #wavfile is  .flac
            sound = AudioSegment.from_wav(wavfile.replace('.flac', self.original_ext)) #wavfile now is  .wav
            sound = sound.set_frame_rate(sr) # downsample it to sr
            sound = sound.set_channels(1) # Convert Stereo to Mono
            sound.export(wavfile.replace(self.original_ext, f'.{output_format}'), format=output_format)   
            #wavfile now is .flac          
        
             
        if num_threads==-1: #when num_threads==-1, it will use all the CPU  
            Parallel(n_jobs=multiprocessing.cpu_count())\
            (delayed(_resample)(wavfile, sr, output_format)\
            for wavfile,tsvfile in tqdm(self._walker,
                                 desc=f'Resampling to {sr}Hz .{output_format} files')) 

        elif num_threads==0:                            
            for wavfile,tsvfile in tqdm(self._walker, desc=f'Resampling to {sr}Hz .{output_format} files'):
                _resample(wavfile, sr, output_format)  #wavfile is the audio path in [self._walker]
        else:
            Parallel(n_jobs=num_threads)\
            (delayed(_resample)(wavfile, sr, output_format)\
            for wavfile, tsvfile in tqdm(self._walker,
                                 desc=f'Resampling to {sr}Hz .{output_format} files'))     

    
            
            


import torch
import numpy as np
import time
from tqdm import tqdm
import hashlib
from mido import Message, MidiFile, MidiTrack
import csv

def tsv2roll(tsv, audio_length, sample_rate, hop_size, max_midi, min_midi):
    """
    Converting a tsv file into a piano roll
    
    Parameters
    ----------
    tsv : numpy.ndarray
        The tsv label in the array format
    hop_size: int
        Hop size for the spectrogram. It will be used to convert tsvs into piano rolls
    max_midi: int
        The top bin of the pianoroll. Default 108 which corresponds to C8, the highest key on a standard 88-key piano
    min_midi: int
        The lowest bin of the pianoroll. Default 21 which corresponds to A0, the lowest key on a standard 88-key piano
    """
    
    
    n_keys = max_midi - min_midi + 1 # Calutate number of bins for the piano roll
    n_steps = (audio_length - 1) // hop_size + 1 # Calulate number of timesteps for the piano roll
    
    pianoroll = torch.zeros((n_steps, n_keys), dtype=int)
    velocity_roll = torch.zeros((n_steps, n_keys), dtype=int)
    
    for onset, offset, note, vel in tsv:
        left = int(round(onset * sample_rate / hop_size)) # Convert time to time step
        onset_right = min(n_steps, left + 1) # Ensure the time step of onset would not exceed the last time step
        frame_right = int(round(offset * sample_rate / hop_size))
        frame_right = min(n_steps, frame_right) # Ensure the time step of frame would not exceed the last time step
        offset_right = min(n_steps, frame_right + 1)

        f = int(note) - min_midi
        pianoroll[left:onset_right, f] = 3 # assigning onset
        pianoroll[onset_right:frame_right, f] = 2 # assigning sustain
        pianoroll[frame_right:offset_right, f] = 1 # assigning offset
        velocity_roll[left:frame_right, f] = vel    
        
        
        
    return pianoroll, velocity_roll


def check_md5(path, md5_hash):
    with open(path, 'rb') as file_to_check:
        # read contents of the file
        data = file_to_check.read()    
        # pipe contents of the file through
        md5_returned = hashlib.md5(data).hexdigest()

        assert md5_returned==md5_hash, f"{os.path.basename(path)} is corrupted, please download it again"

# Helper functions for midi to tsv conversions
def parse_csv(path):
    with open(path, newline='') as csvfile:
        csv_reader = csv.reader(csvfile)
        next(csv_reader, None) # skipping the header
        notes = [] # container for storing tsv entries
        for row in csv_reader:
            # It is fixed to 44100 because the original sr
            onset = int(row[0])/44100 # converting samples to second
            offset = int(row[1])/44100
            pitch = int(row[3])
            velocity = 127
            note = (onset, offset, pitch, velocity)
            notes.append(note)

    return np.array(notes)

def parse_midi(path):
    """open midi file and return np.array of (onset, offset, note, velocity) rows"""
    midi = MidiFile(path)

    time = 0
    sustain = False
    events = []
    for message in midi:
        time += message.time

        if message.type == 'control_change' and message.control == 64 and (message.value >= 64) != sustain:
            # sustain pedal state has just changed
            sustain = message.value >= 64
            event_type = 'sustain_on' if sustain else 'sustain_off'
            event = dict(index=len(events), time=time, type=event_type, note=None, velocity=0)
            events.append(event)

        if 'note' in message.type:
            # MIDI offsets can be either 'note_off' events or 'note_on' with zero velocity
            velocity = message.velocity if message.type == 'note_on' else 0
            event = dict(index=len(events), time=time, type='note', note=message.note, velocity=velocity, sustain=sustain)
            events.append(event)

    notes = []
    for i, onset in enumerate(events):
        if onset['velocity'] == 0:
            continue

        # find the next note_off message
        offset = next(n for n in events[i + 1:] if n['note'] == onset['note'] or n is events[-1])

        if offset['sustain'] and offset is not events[-1]:
            # if the sustain pedal is active at offset, find when the sustain ends
#             offset = next(n for n in events[offset['index'] + 1:] if n['type'] == 'sustain_off' or n is events[-1])
            offset = next(n for n in events[offset['index'] + 1:] if n['type'] == 'sustain_off' or n['note'] == offset['note'] or n is events[-1])
        note = (onset['time'], offset['time'], onset['note'], onset['velocity'])
        notes.append(note)

    return np.array(notes)

def process_csv(input_file, output_file):
    """Parsing CSV files from MusicNet"""
    csv_data = parse_csv(input_file)
    np.savetxt(output_file, csv_data, '%.6f', '\t', header='onset\toffset\tnote\tvelocity')


def process_midi(input_file, output_file):
    midi_data = parse_midi(input_file)
    np.savetxt(output_file, midi_data, '%.6f', '\t', header='onset\toffset\tnote\tvelocity')


def files(file_list, ext='.mid', output_dir=False):
    for input_file in tqdm(file_list, desc='Converting midi to tsv:'):
        if input_file.endswith('.mid') or input_file.endswith('.csv') :
            if output_dir==False:
                output_file = input_file[:-4] + '.tsv'
            else:
                output_file = os.path.join(output_dir, os.path.basename(input_file[:-4]) + '.tsv')
        elif input_file.endswith('.midi'):
            if output_dir==False:
                output_file = input_file[:-5] + '.tsv'
            else:
                output_file = os.path.join(output_dir, os.path.basename(input_file[:-5]) + '.tsv')                
        else:
            print('ignoring non-MIDI file %s' % input_file, file=sys.stderr)
            continue

        yield (input_file, output_file)
        
    


from torch.utils.data import Dataset
import torchaudio
from pathlib import Path
from typing import Tuple, Union
from torch import Tensor
import torch
import os
import time
import tqdm
import shutil
import glob
import multiprocessing as mp
import warnings
from distutils.dir_util import copy_tree
from torchaudio.compliance import kaldi # for downsampling
from torchaudio.datasets.utils import (
    download_url,
    extract_archive,
)
import hashlib
import torch.nn.functional as F            

#start for speechcommands 12 classes code
SAMPLE_RATE = 16000
FOLDER_IN_ARCHIVE = "SpeechCommands"
URL = "speech_commands_v0.02"
HASH_DIVIDER = "_nohash_"
EXCEPT_FOLDER = "_background_noise_"
_CHECKSUMS = {
    "https://storage.googleapis.com/download.tensorflow.org/data/speech_commands_v0.01.tar.gz":
    "3cd23799cb2bbdec517f1cc028f8d43c",
    "https://storage.googleapis.com/download.tensorflow.org/data/speech_commands_v0.02.tar.gz":
    "6b74f3901214cb2c2934e98196829835",
}

UNKNOWN = [
 'backward',
 'bed',
 'bird', 
 'cat', 
 'dog', 
 'eight', 
 'five', 
 'follow', 
 'forward',
 'four', 
 'happy', 
 'house', 
 'learn', 
 'marvin', 
 'nine', 
 'one',  
 'seven', 
 'sheila', 
 'six', 
 'three', 
 'tree', 
 'two', 
 'visual', 
 'wow',
 'zero'
]

name2idx = {
    'down':0,
    'go':1,
    'left':2,
    'no':3,
    'off':4,
    'on':5,
    'right':6,
    'stop':7,
    'up':8,
    'yes':9,
    '_silence_':10,
    '_unknown_':11
}

idx2name = {}
for name, idx in name2idx.items():
    idx2name[idx] = name

def _load_list(root, *filenames):
    output = []
    for filename in filenames:
        filepath = os.path.join(root, filename)
        with open(filepath) as fileobj:
            output += [os.path.normpath(os.path.join(root, line.strip())) for line in fileobj]
    return output


def load_speechcommands_item(filepath: str, path: str) -> Tuple[Tensor, int, str, str, int]:
    relpath = os.path.relpath(filepath, path)
    label, filename = os.path.split(relpath)
    # Besides the officially supported split method for datasets defined by "validation_list.txt"
    # and "testing_list.txt" over "speech_commands_v0.0x.tar.gz" archives, an alternative split
    # method referred to in paragraph 2-3 of Section 7.1, references 13 and 14 of the original
    # paper, and the checksums file from the tensorflow_datasets package [1] is also supported.
    # Some filenames in those "speech_commands_test_set_v0.0x.tar.gz" archives have the form
    # "xxx.wav.wav", so file extensions twice needs to be stripped twice.
    # [1] https://github.com/tensorflow/datasets/blob/master/tensorflow_datasets/url_checksums/speech_commands.txt
    speaker, _ = os.path.splitext(filename)
    speaker, _ = os.path.splitext(speaker)
    
    speaker_id, utterance_number = speaker.split(HASH_DIVIDER)
    utterance_number = int(utterance_number)

    # Load audio
    waveform, sample_rate = torchaudio.load(filepath)
    return waveform, sample_rate, label, speaker_id, utterance_number


def caching_data(_walker, path, subset):
    cache = []
    for filepath in tqdm.tqdm(_walker, desc=f'Loading {subset} set'):
        relpath = os.path.relpath(filepath, path)
        label, filename = os.path.split(relpath)
        if label in UNKNOWN: # if the label is not one of the 10 commands, map them to unknown
            label = '_unknown_'
        
        
        speaker, _ = os.path.splitext(filename)
        speaker, _ = os.path.splitext(speaker)
        
        # When loading test_set, there is a folder for _silence_
        if label == '_silence_':
            speaker_id = speaker.split(HASH_DIVIDER)
            utterance_number = -1
        else:
            speaker_id, utterance_number = speaker.split(HASH_DIVIDER)
            utterance_number = int(utterance_number)

        # Load audio     
        audio_samples, rate = torchaudio.load(filepath) # loading audio
        # audio_sample (1, len)
        
        if audio_samples.shape[1] != SAMPLE_RATE:
            pad_length = SAMPLE_RATE-audio_samples.shape[1]
            audio_samples = F.pad(audio_samples, (0,pad_length)) # pad the end of the audio until 1 second
            # (1, 16000)
        cache.append((audio_samples, rate, name2idx[label], speaker_id, utterance_number)) 
    
    
    # include silence
    if subset=='training':
        slience_clips = [
            'dude_miaowing.wav',
            'white_noise.wav',
            'exercise_bike.wav',
            'doing_the_dishes.wav',
            'pink_noise.wav'
        ]
    elif subset=='validation':
        slience_clips = [
            'running_tap.wav'
        ]
    else:
        slience_clips = []
        
        
    for i in slience_clips: 
        audio_samples, rate = torchaudio.load(os.path.join(path, '_background_noise_', i))
        for start in range(0,
                           audio_samples.shape[1] - SAMPLE_RATE,
                           SAMPLE_RATE//2):
            audio_segment = audio_samples[0, start:start + SAMPLE_RATE]
            cache.append((audio_segment.unsqueeze(0), rate, name2idx['_silence_'], '00000000', -1))        
        
    return cache

class SPEECHCOMMANDS_12C(Dataset):
    """Create a Dataset for Speech Commands.
    Args:
        root (str or Path): Path to the directory where the dataset is found or downloaded.
        url (str, optional): The URL to download the dataset from,
            or the type of the dataset to dowload.
            Allowed type values are ``"speech_commands_v0.01"`` and ``"speech_commands_v0.02"``
            (default: ``"speech_commands_v0.02"``)
        folder_in_archive (str, optional):
            The top-level directory of the dataset. (default: ``"SpeechCommands"``)
        download (bool, optional):
            Whether to download the dataset if it is not found at root path. (default: ``False``).
        subset (str or None, optional):
            Select a subset of the dataset [None, "training", "validation", "testing"]. None means
            the whole dataset. "validation" and "testing" are defined in "validation_list.txt" and
            "testing_list.txt", respectively, and "training" is the rest. Details for the files
            "validation_list.txt" and "testing_list.txt" are explained in the README of the dataset
            and in the introduction of Section 7 of the original paper and its reference 12. The
            original paper can be found `here <https://arxiv.org/pdf/1804.03209.pdf>`_. (Default: ``None``)
    """

    def __init__(self,
                 root,
                 url,
                 folder_in_archive,
                 download,
                 subset,
                 ):

        assert subset is None or subset in ["training", "validation", "testing"], (
            "When `subset` not None, it must take a value from "
            + "{'training', 'validation', 'testing'}."
        )

        if subset in ["training", "validation"]:
            url = "speech_commands_v0.02"
            
        elif subset=='testing':
            url = "speech_commands_test_set_v0.02"
        
        base_url = "https://storage.googleapis.com/download.tensorflow.org/data/"
        ext_archive = ".tar.gz"

        url = os.path.join(base_url, url + ext_archive)

        # Get string representation of 'root' in case Path object is passed
        root = os.fspath(root)

        basename = os.path.basename(url)
        print(f"{basename=}")
        archive = os.path.join(root, basename)

        basename = basename.rsplit(".", 2)[0]
        folder_in_archive = os.path.join(folder_in_archive, basename)

        self._path = os.path.join(root, folder_in_archive)

        if download:
            if not os.path.isdir(self._path):
                if not os.path.isfile(archive):
                    checksum = _CHECKSUMS.get(url, None)
                    download_url(url, root, hash_value=checksum, hash_type="md5")
                extract_archive(archive, self._path)

        if subset == "validation":
            self._walker = _load_list(self._path, "validation_list.txt")
            self._data = caching_data(self._walker, self._path, subset)            
        elif subset == "testing":
            self._walker = list(Path(self._path).glob('*/*.wav'))
            self._data = caching_data(self._walker, self._path, subset)
        elif subset == "training":
            excludes = set(_load_list(self._path, "validation_list.txt", "testing_list.txt"))
            walker = sorted(str(p) for p in Path(self._path).glob('*/*.wav'))
            self._walker = [
                w for w in walker
                if HASH_DIVIDER in w
                and EXCEPT_FOLDER not in w
                and os.path.normpath(w) not in excludes
            ]
            self._data = caching_data(self._walker, self._path, subset)
            
        else:
            walker = sorted(str(p) for p in Path(self._path).glob('*/*.wav'))
            self._walker = [w for w in walker if HASH_DIVIDER in w and EXCEPT_FOLDER not in w]

    def __getitem__(self, n: int) -> Tuple[Tensor, int, str, str, int]:
        """Load the n-th sample from the dataset.
        Args:
            n (int): The index of the sample to be loaded
        Returns:
            (Tensor, int, str, str, int):
            ``(waveform, sample_rate, label, speaker_id, utterance_number)``
        """
        return self._data[n]


    def __len__(self) -> int:
        return len(self._data)       







import setuptools
import codecs
import os.path

with open("README.md", "r") as fh:
    long_description = fh.read()

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()    
    
def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")    
    
setuptools.setup(
    name="AudioLoader", # Replace with your own username
    version=get_version("AudioLoader/__init__.py"),
    author="KinWaiCheuk",
    author_email="u3500684@connect.hku.hk",
    description="A collection of PyTorch audio datasets for speech and music applications",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/KinWaiCheuk/AudioLoader",
    packages=setuptools.find_packages(),
    include_package_data=True,    
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    pytorch='>=1.6'
)


import json
import os
from pathlib import Path
from abc import abstractmethod
from glob import glob
import csv
import shutil
import sys
import pickle
import numpy as np
import random
import yaml
from typing import Optional, Callable
# import soundfile
from tqdm import tqdm
import multiprocessing
import musdb
import torch
import torch.nn.functional as F
from torch.utils.data import Dataset
import torchaudio
from torchaudio.compliance import kaldi # for downsampling
import hashlib
from torchaudio.datasets.utils import (
    download_url,
    extract_archive,
)

from collections import OrderedDict
import math

metadata= './metadata'
MIXTURE = 'mixture'
EXT = '.wav'

class MusdbHQ:
    def __init__(
            self,root, subset, sources, download = False,segment=None, shift=None, normalize=True,
            samplerate=44100, channels=2, ext=EXT):
        """
        MusdbHQ (or mp3 set for that matter). Can be used to train
        with arbitrary sources. Each track should be one folder inside of `path`.
        The folder should contain files named `{source}.{ext}`.
        Args:
            root (Path or str): root folder for the dataset.
            subset (str): training ,validation, traininf_all or test
            sources (list): ['drums', 'bass', 'other', 'vocals']
            download (bool): Whether to download the dataset if it is not found at root path. (default: ``False``).
            segment (None or float): segment length in seconds. If `None`, returns entire tracks.
            shift (None or float): stride in seconds bewteen samples.
            normalize (bool): normalizes input audio, **based on the metadata content**,
                i.e. the entire track is normalized, not individual extracts.
            samplerate (int): target sample rate. if the file sample rate
                is different, it will be resampled on the fly.
            channels (int): target nb of channels. if different, will be
                changed onthe fly.
            ext (str): extension for audio files (default is .wav).
        samplerate and channels are converted on the fly.
        """
 
         # Getting audio path
        ext_archive = '.zip'
        archive_name = 'musdb18hq'

        download_path = Path(root)
        self.download_path = download_path
               
        if subset.lower() == 'training' or subset.lower() == 'validation' or subset.lower() == 'training_all':
            dataset = 'train'
        elif subset.lower() == 'test':
            dataset = 'test'
        else:
            print(f'Subset does not exist, please choose from training, validation, training_all or test')
        
    
        self._path = os.path.join(root, dataset.lower())

        url= 'https://zenodo.org/record/3338373/files/musdb18hq.zip?download=1'
        checksum = '12d4f2ecd55245a4688754dd76363103'
        
        if download:
            if os.path.isdir(download_path) and os.path.isdir(os.path.join(download_path, 'train')) and os.path.isdir(os.path.join(download_path, 'test')):
                print(f'Dataset folder exists, skipping download...')
                decision = input(f"Do you want to extract {archive_name+ext_archive} again? "
                                 f"To avoid this prompt, set `download=False`\n"
                                 f"This action will overwrite exsiting files, do you still want to continue? [yes/no]") 
                if decision.lower()=='yes':
                    print(f'extracting...')
                    extract_archive(os.path.join(download_path, archive_name+ext_archive))                
            elif os.path.isfile(os.path.join(download_path, 'musdb18hq.zip')):
                print(f'musdb18hq.zip exists, checking MD5...')
                check_md5(os.path.join(download_path, archive_name+ext_archive), checksum)
                print(f'MD5 is correct, extracting...')
                extract_archive(os.path.join(download_path, archive_name+ext_archive))
            else:
                decision='yes'       
                if not os.path.isdir(download_path):
                    print(f'Creating download path = {root}')
                    os.makedirs(os.path.join(download_path))

                if os.path.isfile(os.path.join(download_path, archive_name+ext_archive)):
                    print(f'{download_path+archive_name+ext_archive} already exists, proceed to extraction...')
                else:
                    print(f'downloading...')
                    try:
                        download_url(url, download_path, hash_value=checksum, hash_type='md5')
                    except:
                        raise Exception('Auto download fails. '+
                                        'You may want to download it manually from:\n'+
                                        url+ '\n' +
                                        f'Then, put it inside {download_path}')

        
        if os.path.isdir(self._path):
            pass
        elif os.path.isfile(os.path.join(download_path, 'musdb18hq.zip')):
            print(f'musdb18hq.zip exists, checking MD5...')
            check_md5(os.path.join(download_path, archive_name+ext_archive), checksum)
            print(f'MD5 is correct, extracting...')            
            extract_archive(os.path.join(download_path, archive_name+ext_archive))
            
        else:
            raise FileNotFoundError(f"Dataset not found at {self._path}, please specify the correct location or set `download=True`")
                    
        print(f'Using data at {self._path}')
        # It seems we don't need to + ._ext_audio
        self._walker = []
                
# get_musdb_wav_datasets             
        if subset.lower() == 'training' or subset.lower() == 'validation' or subset.lower() == 'training_all':
            root = Path(root) / "train"
            metadata_file = Path('./metadata') / ('musdb_' + 'train' + ".json")
        
        elif subset.lower() == 'test':
            root = Path(root) / "test"
            metadata_file = Path('./metadata') / ('musdb_' + 'test' + ".json")
        
    #     if not metadata_file.is_file() and distrib.rank == 0:
        if not metadata_file.is_file():
            metadata_file.parent.mkdir(exist_ok=True, parents=True)
            metadata = build_metadata(root, sources)
            json.dump(metadata, open(metadata_file, "w"))
    #     if distrib.world_size > 1:
    #         distributed.barrier()
        metadata = json.load(open(metadata_file))

        valid_tracks = _get_musdb_valid()
        
        if subset.lower() == 'training':
            metadata = {name: meta for name, meta in metadata.items() if name not in valid_tracks}
            self.sources = sources
            
        elif subset.lower() == 'validation':
            metadata = {name: meta for name, meta in metadata.items() if name in valid_tracks}
            self.sources = [MIXTURE] + list(sources)
            
        elif subset.lower() == 'training_all':
            metadata_train = metadata
            self.sources = sources
        
        elif subset.lower() == 'test':
            metadata_test = metadata
            self.sources = [MIXTURE] + list(sources)    
            
# metadata (dict): output from `build_metadata`.
# sources (list[str]): list of source names.
        
        self.root = Path(root)
        self.metadata = OrderedDict(metadata)
        self.segment = segment
        self.shift = shift or segment
        self.normalize = normalize
        self.channels = channels
        self.samplerate = samplerate
        self.ext = ext
        self.num_examples = []
        for name, meta in self.metadata.items():
            track_duration = meta['length'] / meta['samplerate']
            if segment is None or track_duration < segment:
                examples = 1
            else:
                examples = int(math.ceil((track_duration - self.segment) / self.shift) + 1)
            self.num_examples.append(examples)
# samplerate = number of sample per second
# length = number of sample
# track_duration is second, cut song into segment

    def __len__(self):
        return sum(self.num_examples)

    def get_file(self, name, source):
        return self.root / name / f"{source}{self.ext}"

    def __getitem__(self, index):
#         print(len(self.metadata))
        for name, examples in zip(self.metadata, self.num_examples):           
            if index >= examples:
                index -= examples
                continue
            meta = self.metadata[name]
            num_frames = -1
            offset = 0
            if self.segment is not None:
                offset = int(meta['samplerate'] * self.shift * index)
                num_frames = int(math.ceil(meta['samplerate'] * self.segment))
            wavs = []
            for source in self.sources:
                file = self.get_file(name, source)
                wav, _ = torchaudio.load(str(file), frame_offset=offset, num_frames=num_frames)
                wav = convert_audio_channels(wav, self.channels)
                wavs.append(wav)

            example = torch.stack(wavs)
            example = torchaudio.functional.resample(example, meta['samplerate'], self.samplerate)
            if self.normalize:
                example = (example - meta['mean']) / meta['std']
            if self.segment:
                length = int(self.segment * self.samplerate)
                example = example[..., :length]
                example = F.pad(example, (0, length - example.shape[-1]))
            return example


def convert_audio_channels(wav, channels=2):
    """Convert audio to the given number of channels."""
    *shape, src_channels, length = wav.shape
    if src_channels == channels:
        pass
    elif channels == 1:
        # Case 1:
        # The caller asked 1-channel audio, but the stream have multiple
        # channels, downmix all channels.
        wav = wav.mean(dim=-2, keepdim=True)
    elif src_channels == 1:
        # Case 2:
        # The caller asked for multiple channels, but the input file have
        # one single channel, replicate the audio over all channels.
        wav = wav.expand(*shape, channels, length)
    elif src_channels >= channels:
        # Case 3:
        # The caller asked for multiple channels, and the input file have
        # more channels than requested. In that case return the first channels.
        wav = wav[..., :channels, :]
    else:
        # Case 4: What is a reasonable choice here?
        raise ValueError('The audio file has less channels than requested but is not mono.')
    return wav


def build_metadata(path, sources, normalize=True, ext=EXT):
    """
    Build the metadata for `MusdbHQ`.
    Args:
        path (str or Path): path to dataset.
        sources (list[str]): list of sources to look for.
        normalize (bool): if True, loads full track and store normalization
            values based on the mixture file.
        ext (str): extension of audio files (default is .wav).
    """

    meta = {}
    path = Path(path)
    pendings = []
    from concurrent.futures import ThreadPoolExecutor
    with ThreadPoolExecutor(8) as pool:
        for root, folders, files in os.walk(path, followlinks=True):
            root = Path(root)
            if root.name.startswith('.') or folders or root == path:
                continue
            name = str(root.relative_to(path))
            pendings.append((name, pool.submit(_track_metadata, root, sources, normalize, ext)))
            # meta[name] = _track_metadata(root, sources, normalize, ext)
        for name, pending in tqdm(pendings, ncols=120):
            meta[name] = pending.result()
    return meta



def _track_metadata(track, sources, normalize=True, ext=EXT):
    track_length = None
    track_samplerate = None
    mean = 0
    std = 1
    for source in sources + [MIXTURE]:
        file = track / f"{source}{ext}"
        try:
            info = torchaudio.info(str(file))
        except RuntimeError:
            print(file)
            raise
        length = info.num_frames
        if track_length is None:
            track_length = length
            track_samplerate = info.sample_rate
        elif track_length != length:
            raise ValueError(
                f"Invalid length for file {file}: "
                f"expecting {track_length} but got {length}.")
        elif info.sample_rate != track_samplerate:
            raise ValueError(
                f"Invalid sample rate for file {file}: "
                f"expecting {track_samplerate} but got {info.sample_rate}.")
        if source == MIXTURE and normalize:
            try:
                wav, _ = torchaudio.load(str(file))
            except RuntimeError:
                print(file)
                raise
            wav = wav.mean(0)
            mean = wav.mean().item()
            std = wav.std().item()

    return {"length": length, "mean": mean, "std": std, "samplerate": track_samplerate}


def _get_musdb_valid():
    # Return musdb valid set.
    import yaml
    setup_path = Path(musdb.__path__[0]) / 'configs' / 'mus.yaml'
    setup = yaml.safe_load(open(setup_path, 'r'))
    return setup['validation_tracks']
        
        
def check_md5(path, md5_hash):
    """
    This version of cehck_md5 reads file chunk by chunk to avoid memory error (file size > RAM)
    """
    md5 = hashlib.md5()
    with open(path,'rb') as f: 
        for chunk in iter(lambda: f.read(8192), b''): 
            md5.update(chunk)
    md5_returned = md5.hexdigest()
    assert md5_returned==md5_hash, f"{os.path.basename(path)} is corrupted, please download it again"
